@@ -55,10 +55,9 @@ typedef unsigned int NSUInteger;
 #endif
 #endif
 
-#if MAC_OS_X_VERSION_MIN_REQUIRED >= 1060
-@interface X11Controller : NSObject <NSTableViewDataSource>
-#else
 @interface X11Controller : NSObject
+#if MAC_OS_X_VERSION_MIN_REQUIRED >= 1060
+<NSTableViewDataSource>
 #endif
 {
     IBOutlet NSPanel *prefs_panel;
@@ -69,6 +68,7 @@ typedef unsigned int NSUInteger;
     IBOutlet NSButton *use_sysbeep;
     IBOutlet NSButton *enable_keyequivs;
     IBOutlet NSButton *sync_keymap;
+    IBOutlet NSButton *option_sends_alt;
     IBOutlet NSButton *click_through;
     IBOutlet NSButton *focus_follows_mouse;
     IBOutlet NSButton *focus_on_new_window;
